import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10
from libtaxii.common import generate_message_id

from libtaxii.constants import (
    VID_TAXII_XML_11, VID_TAXII_XML_10,
    VID_TAXII_SERVICES_10, VID_TAXII_SERVICES_11
)

from .http import (
    HTTP_X_TAXII_ACCEPT, HTTP_X_TAXII_CONTENT_TYPE,
    get_http_headers
)


def process_status_exception(exception, headers, is_secure):

    accepted_content = headers.get(HTTP_X_TAXII_ACCEPT)

    if not accepted_content:  # Can respond with whatever we want. try to use the X-TAXII-Content-Type header to pick
        accepted_content = headers.get(HTTP_X_TAXII_CONTENT_TYPE)

    try:
        status = exception_to_status(exception, accepted_content)
    except ValueError:
        status = exception_to_status(exception, VID_TAXII_XML_11)

    if accepted_content == VID_TAXII_XML_10:
        version = VID_TAXII_SERVICES_10
    elif accepted_content == VID_TAXII_XML_11:
        version = VID_TAXII_SERVICES_11
    else:
        #FIXME: Unknown accepted content. Pretending X-TAXII-Accept was TAXII 1.1
        version = VID_TAXII_SERVICES_11

    response_headers = get_http_headers(version, is_secure)

    return status.to_xml(pretty_print=True), response_headers


def exception_to_status(exception, format_version):

    data = dict(
        message_id = generate_message_id(),
        in_response_to = exception.in_response_to,
        extended_headers = exception.extended_headers,
        status_type = exception.status_type,
        status_detail = exception.status_details,
        message = exception.message
    )

    if format_version == VID_TAXII_XML_11:
        sm = tm11.StatusMessage(**data)
    elif format_version == VID_TAXII_XML_10:
        sm = tm10.StatusMessage(**data)
    else:
        raise ValueError("Unknown version: %s" % format_version)

    return sm






import argparse
import structlog

from opentaxii.server import TAXIIServer
from opentaxii.config import ServerConfig
from opentaxii.utils import configure_logging

config = ServerConfig()
configure_logging(config.get('logging'), plain=True)

log = structlog.getLogger(__name__)


def get_parser():
    parser = argparse.ArgumentParser(
        description = "Create Account via OpenTAXII Auth API",
        formatter_class = argparse.ArgumentDefaultsHelpFormatter
    )
    return parser


def create_account():

    parser = get_parser()
    parser.add_argument("-u", "--username", dest="username", help="Username for the new account", required=True)
    parser.add_argument("-p", "--password", dest="password", help="Password for the new account", required=True)

    args = parser.parse_args()

    server = TAXIIServer(config)
    token = server.auth.create_account(args.username, args.password)

    log.info("Account created", token=token)




import sys
import logging
import structlog
import urlparse
import importlib
import base64

from .config import ServerConfig
from .exceptions import InvalidAuthHeader

log = structlog.getLogger(__name__)


def get_path_and_address(domain, address):
    parsed = urlparse.urlparse(address)

    if parsed.scheme:
        return None, address
    else:
        return address, domain + address


def import_class(module_class_name):
    module_name, _, class_name = module_class_name.rpartition('.')
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def load_api(api_config):
    cls = import_class(api_config['class'])
    params = api_config['parameters']

    if params:
        instance = cls(**params)
    else:
        instance = cls()
    return instance


def parse_basic_auth_token(token):
    try:
        value = base64.b64decode(token)
    except TypeError:
        raise InvalidAuthHeader("Can't decode Basic Auth header value")

    try:
        username, password = value.split(':', 1)
        return (username, password)
    except ValueError:
        raise InvalidAuthHeader("Invalid Basic Auth header value")


class PlainRenderer(object):

    def __call__(self, logger, name, event_dict):
        pairs = ', '.join(['%s=%s' % (k, v) for k, v in event_dict.items()])
        return '%(timestamp)s [%(logger)s] %(level)s: %(event)s {%(pairs)s}' \
                % dict(pairs=pairs, **event_dict)


def configure_logging(logging_levels, plain=False):

    _remove_all_existing_log_handlers()

    renderer = PlainRenderer() if plain \
            else structlog.processors.JSONRenderer()

    structlog.configure(
        processors = [
            structlog.stdlib.filter_by_level,
            structlog.stdlib.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.TimeStamper(fmt='iso'),
            structlog.processors.StackInfoRenderer(),
            structlog.processors.format_exc_info,
            renderer
        ],
        context_class = dict,
        logger_factory = structlog.stdlib.LoggerFactory(),
        wrapper_class = structlog.stdlib.BoundLogger,
        cache_logger_on_first_use = True,
    )

    handler = logging.StreamHandler(sys.stdout)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)

    for logger, level in logging_levels.items():

        if logger.lower() == 'root':
            logger = ''

        logging.getLogger(logger).setLevel(level.upper())


def _remove_all_existing_log_handlers():
    for logger in logging.Logger.manager.loggerDict.values():
        if hasattr(logger, 'handlers'):
            del logger.handlers[:]

    root_logger = logging.getLogger()
    del root_logger.handlers[:]


def get_config_for_tests(domain, persistence_db=None, auth_db=None):

    config = ServerConfig()
    config.update({
        'persistence_api' : {
            'class' : 'opentaxii.persistence.sqldb.SQLDatabaseAPI',
            'parameters' : {
                'db_connection' : persistence_db or 'sqlite://',
                'create_tables' : True
            }
        },
        'auth_api' : {
            'class' : 'opentaxii.auth.sqldb.SQLDatabaseAPI',
            'parameters' : {
                'db_connection' : auth_db or 'sqlite://',
                'create_tables' : True,
                'secret' : 'dummy-secret-string-for-tests'
            }
        }
    })
    config['domain'] = domain
    return config





import pytz
import structlog
from datetime import datetime

from lxml.etree import XMLSyntaxError

from .exceptions import BadMessageStatus
from .bindings import MESSAGE_VALIDATOR_PARSER

log = structlog.getLogger(__name__)


def get_utc_now():
    return datetime.utcnow().replace(tzinfo=pytz.UTC)


def is_content_supported(supported_bindings, content_binding, version=None):

    if not hasattr(content_binding, 'binding_id') or version == 10:
        binding_id = content_binding
        subtype = None
    else:
        binding_id = content_binding.binding_id

        #FIXME: may be not the best option
        subtype = content_binding.subtype_ids[0] if content_binding.subtype_ids \
                else None

    matches = [
        ((supported.binding == binding_id) and (not supported.subtypes or subtype in supported.subtypes))
        for supported in supported_bindings
    ]

    return any(matches)


def parse_message(content_type, body, do_validate=True):

    validator_parser = MESSAGE_VALIDATOR_PARSER[content_type]

    if do_validate:
        try:
            result = validator_parser.validator.validate_string(body)
            if not result.valid:
                errors = '; '.join([str(err) for err in result.error_log])
                raise BadMessageStatus('Request was not schema valid: %s' % errors)
        except XMLSyntaxError as e:
            log.error("Invalid XML received", exc_info=True)
            raise BadMessageStatus('Request was invalid XML', e=e)

    taxii_message = validator_parser.parser(body)

    return taxii_message





import json
import structlog
from sqlalchemy import orm, engine, func, and_, or_

from opentaxii.persistence import OpenTAXIIPersistenceAPI

from . import models
from . import converters as conv

__all__ = ['SQLDatabaseAPI']

log = structlog.getLogger(__name__)

YIELD_PER_SIZE = 100


class SQLDatabaseAPI(OpenTAXIIPersistenceAPI):
    """SQL database implementation of OpenTAXII Persistence API.

    Implementation will work with any DB supported by SQLAlchemy package.

    Note: this implementation ignores ``context.account`` and does not have
    any access rules.

    :param str db_connection: a string that indicates database dialect and
                          connection arguments that will be passed directly
                          to :func:`~sqlalchemy.engine.create_engine` method.

    :param bool create_tables=False: if True, tables will be created in the DB.
    """

    def __init__(self, db_connection, create_tables=False):

        self.engine = engine.create_engine(db_connection, convert_unicode=True)

        self.Session = orm.scoped_session(orm.sessionmaker(autocommit=False,
            autoflush=True, bind=self.engine))

        attach_all(self, models)

        self.Base.query = self.Session.query_property()

        if create_tables:
            self.create_tables()

    def create_tables(self):
        self.Base.metadata.create_all(bind=self.engine)

    def _merge(self, obj):
        s = self.Session()
        updated = s.merge(obj)
        s.commit()
        return updated

    def get_services(self, collection_id=None):
        if collection_id:
            collection = self.DataCollection.query.get(collection_id)
            services = collection.services
        else:
            services = self.Service.query.all()
        return map(conv.to_service_entity, services)

    def get_service(self, service_id):
        return conv.to_service_entity(self.Service.get(service_id))

    def update_service(self, entity):
        service = self.Service(id=entity.id, type=entity.type,
                properties=entity.properties)
        updated = self._merge(service)
        return conv.to_service_entity(updated)

    def create_service(self, entity):
        return self.update_service(entity)

    def get_collections(self, service_id):
        service = self.Service.query.get(service_id)
        return map(conv.to_collection_entity, service.collections)

    def get_collection(self, name, service_id):

        collection = (self.DataCollection.query
                                         .join(self.Service.collections)
                                         .filter(self.Service.id == service_id)
                                         .filter_by(name=name)).first()
        if collection:
            return conv.to_collection_entity(collection)

    def _get_content_query(self, collection_id=None, start_time=None,
            end_time=None, bindings=None, count=False):

        if count:
            query = self.Session().query(func.count(self.ContentBlock.id))
        else:
            query = self.ContentBlock.query

        if collection_id:
            query = (query.join(self.ContentBlock.collections)
                          .filter(self.DataCollection.id == collection_id))

        if start_time:
            query = query.filter(self.ContentBlock.date_created > start_time)

        if end_time:
            query = query.filter(self.ContentBlock.date_created <= end_time)

        if bindings:
            criteria = []
            for binding in bindings:
                if binding.subtypes:
                    criterion = and_(
                       self.ContentBlock.binding_id == binding.binding,
                       self.ContentBlock.binding_subtype.in_(binding.subtypes)
                    )
                else:
                    criterion = self.ContentBlock.binding_id == binding.binding
                criteria.append(criterion)

            query = query.filter(or_(*criteria))

        return query

    def get_content_blocks_count(self, collection_id=None, start_time=None,
            end_time=None, bindings=None):

        query = self._get_content_query(collection_id=collection_id,
                start_time=start_time, end_time=end_time,
                bindings=bindings, count=True)

        return query.scalar()

    def get_content_blocks(self, collection_id=None, start_time=None,
            end_time=None, bindings=None, offset=0, limit=None):

        query = self._get_content_query(collection_id=collection_id,
                start_time=start_time, end_time=end_time,
                bindings=bindings)

        query = query.offset(offset)
        if limit:
            query = query.limit(limit)

        blocks = query.yield_per(YIELD_PER_SIZE)

        return map(conv.to_block_entity, query.all())

    def update_collection(self, entity):

        _bindings = conv.serialize_content_bindings(entity.supported_content)

        collection = self.DataCollection(
            id = entity.id,
            name = entity.name,
            type = entity.type,
            description = entity.description,
            available = entity.available,
            accept_all_content = entity.accept_all_content,
            bindings = _bindings
        )

        updated = self._merge(collection)

        log.debug("Collection updated", collection_id=updated.id,
                collection_name=updated.name)

        return conv.to_collection_entity(updated)

    def create_collection(self, entity):
        return self.update_collection(entity)

    def attach_collection_to_services(self, collection_id, service_ids):
        
        collection = self.DataCollection.query.get(collection_id)

        if not collection:
            raise ValueError("Can't find collection with id=%s" % collection_id)

        s = self.Session()

        collection.services.extend(self.Service.query.filter(
            self.Service.id.in_(service_ids)))

        s.add(collection)
        s.commit()

        log.debug("Collection attached", collection_id=collection.id,
                collection_name=collection.name, service_ids=service_ids)

    def create_inbox_message(self, entity):
        return self.update_inbox_message(entity)

    def update_inbox_message(self, entity):

        if entity.destination_collections:
            names = json.dumps(entity.destination_collections) 
        else:
            names = None

        message = self.InboxMessage(
            id = entity.id,
            message_id = entity.message_id,
            original_message = entity.original_message,
            content_block_count = entity.content_block_count,
            destination_collections = names,

            service_id = entity.service_id,

            result_id = entity.result_id,
            record_count = entity.record_count,
            partial_count = entity.partial_count,

            subscription_collection_name = entity.subscription_collection_name,
            subscription_id = entity.subscription_id,

            exclusive_begin_timestamp_label = entity.exclusive_begin_timestamp_label,
            inclusive_end_timestamp_label = entity.inclusive_end_timestamp_label,
        )

        updated = self._merge(message)
        return conv.to_inbox_message_entity(updated)

    def update_content_block(self, entity):

        if entity.content_binding:
            binding = entity.content_binding.binding
            subtype = entity.content_binding.subtypes[0] \
                    if entity.content_binding.subtypes else None
        else:
            binding = None
            subtype = None

        content = self.ContentBlock(
            id = entity.id,
            timestamp_label = entity.timestamp_label,
            inbox_message_id = entity.inbox_message_id,
            content = entity.content,
            binding_id = binding,
            binding_subtype = subtype
        )

        updated = self._merge(content)
        return conv.to_block_entity(updated)

    def create_content_block(self, entity, collection_ids=None,
            service_id=None):

        block = self.update_content_block(entity)
        if collection_ids:
            self._attach_content_to_collections(block, collection_ids)

        return block

    def _attach_content_to_collections(self, content_block, collection_ids):

        if not collection_ids:
            return

        session = self.Session()
        criteria = self.DataCollection.id.in_(collection_ids)
        new_collections = self.DataCollection.query.filter(criteria)

        content_block = self.ContentBlock.query.get(content_block.id)
        content_block.collections.extend(new_collections)
        session.add(content_block)

        log.debug("Content block added to collections",
                  content_block_id=content_block.id,
                  collections=new_collections.count())
        session.commit()

    def update_result_set(self, entity):

        _bindings = conv.serialize_content_bindings(entity.content_bindings)

        result_set = self.ResultSet(
            id = entity.result_id,
            collection_id = entity.collection_id,
            bindings = _bindings,
            begin_time = entity.timeframe[0],
            end_time = entity.timeframe[1]
        )

        updated = self._merge(result_set)
        return conv.to_result_set_entity(updated)

    def create_result_set(self, entity):
        return self.update_result_set(entity)

    def get_result_set(self, result_set_id):
        result_set = self.ResultSet.query.get(result_set_id)
        return conv.to_result_set_entity(result_set)

    def get_subscription(self, subscription_id):
        s = self.Subscription.query.get(subscription_id)
        return conv.to_subscription_entity(s)

    def get_subscriptions(self, service_id):
        service = self.Service.query.get(service_id)
        return map(conv.to_subscription_entity, service.subscriptions)

    def update_subscription(self, entity):

        if entity.params:
            params = dict(
                response_type = entity.params.response_type,
                content_bindings = conv.serialize_content_bindings(
                        entity.params.content_bindings)
            )
        else:
            params = {}

        subscription = self.Subscription(
            id = entity.subscription_id,
            collection_id = entity.collection_id,
            params = json.dumps(params),
            status = entity.status,
            service_id = entity.service_id
        )

        updated = self._merge(subscription)
        log.debug("Subscription updated", subscription_id=updated.id,
                collection_id=updated.collection_id, status=updated.status)

        return conv.to_subscription_entity(updated)

    def create_subscription(self, entity):
        return self.update_subscription(entity)


def attach_all(obj, module):
    for key in module.__all__:
        if not hasattr(obj, key):
            setattr(obj, key, getattr(module, key))
    return obj



from libtaxii.constants import (
        SVC_COLLECTION_MANAGEMENT,
        MSG_COLLECTION_INFORMATION_REQUEST, MSG_FEED_INFORMATION_REQUEST,
        MSG_MANAGE_COLLECTION_SUBSCRIPTION_REQUEST,
        MSG_MANAGE_FEED_SUBSCRIPTION_REQUEST,
)

from .abstract import TAXIIService
from .handlers import (
        CollectionInformationRequestHandler,
        SubscriptionRequestHandler
)


class CollectionManagementService(TAXIIService):

    handlers = {
        MSG_COLLECTION_INFORMATION_REQUEST : CollectionInformationRequestHandler,
        MSG_FEED_INFORMATION_REQUEST : CollectionInformationRequestHandler,

    }

    subscription_handlers = {
        MSG_MANAGE_COLLECTION_SUBSCRIPTION_REQUEST : SubscriptionRequestHandler,
        MSG_MANAGE_FEED_SUBSCRIPTION_REQUEST : SubscriptionRequestHandler
    }
            
    service_type = SVC_COLLECTION_MANAGEMENT

    subscription_message = "Default subscription message"
    subscription_supported = True

    def __init__(self, subscription_supported=True, subscription_message=None, **kwargs):
        super(CollectionManagementService, self).__init__(**kwargs)

        self.subscription_message = subscription_message
        self.subscription_supported = subscription_supported

        if self.subscription_supported:
            self.handlers = dict(CollectionManagementService.handlers)
            self.handlers.update(CollectionManagementService.subscription_handlers)


    @property
    def advertised_collections(self):
        return self.server.persistence.get_collections(self.id)


    def get_collection(self, name):
        return self.server.persistence.get_collection(name, self.id)


    def get_push_methods(self, collection):
        # Push delivery is not implemented
        pass


    def get_polling_services(self, collection):
        return self.server.get_services_for_collection(collection, 'poll')

    def get_subscription_services(self, collection):
        services = []
        all_services = self.server.get_services_for_collection(collection,
                'collection_management')
        for s in all_services:
            if s.subscription_supported:
                services.append(s)
        return services

    def create_subscription(self, subscription):
        subscription.subscription_id = self.generate_id()
        return self.server.persistence.create_subscription(subscription)

    def get_subscription(self, subscription_id):
        return self.server.persistence.get_subscription(subscription_id)

    def get_subscriptions(self):
        return self.server.persistence.get_subscriptions(service_id=self.id)

    def update_subscription(self, subscription):
        return self.server.persistence.update_subscription(subscription)

    def get_receiving_inbox_services(self, collection):
        return self.server.get_services_for_collection(collection, 'inbox')

    def get_volume(self, collection):
        return self.server.persistence.get_content_blocks_count(collection.id)





from libtaxii.constants import SVC_DISCOVERY, MSG_DISCOVERY_REQUEST

from .abstract import TAXIIService
from .handlers import DiscoveryRequestHandler

class DiscoveryService(TAXIIService):

    service_type = SVC_DISCOVERY

    handlers = {
        MSG_DISCOVERY_REQUEST : DiscoveryRequestHandler
    }

    advertised_services = []

    def __init__(self, services=None, **kwargs):
        super(DiscoveryService, self).__init__(**kwargs)

        self.advertised_services = services or []


    def set_advertised_services(self, services):
        self.advertised_services = services





import structlog

from libtaxii.common import generate_message_id
from libtaxii.constants import (
    VID_TAXII_XML_10, VID_TAXII_XML_11
)

from ..exceptions import StatusMessageException, raise_failure
from ..bindings import PROTOCOL_TO_SCHEME
from ..converters import service_to_service_instances


class TAXIIService(object):
    '''Generic TAXII Service class.

    This class implements common methods for all TAXII services.

    :param str id: service ID
    :param str address: service address as absolute URL
    :param str description: service description

    :param list protocol_bindings: list of supported protocol bindings
        as a list of strings
    :param bool available: if the service is available
    :param bool authentication_required: if authentication required
    :param str path: relative path if service is configured in the server
    '''

    id = None
    description = 'Default TAXII service description'
    available = True

    server = None

    authentication_required = False

    supported_message_bindings = [VID_TAXII_XML_10, VID_TAXII_XML_11]
    supported_protocol_bindings = ()

    def __init__(self, id, server, address, description=None, path=None,
            protocol_bindings=None, available=True, authentication_required=False):

        self.id = id
        self.server = server

        self.address = address
        self.path = path

        self.description = description
        self.supported_protocol_bindings = protocol_bindings or self.supported_protocol_bindings

        self.available = available
        self.authentication_required = authentication_required

        self.log = structlog.getLogger("%s.%s" % (self.__module__,
            self.__class__.__name__), service_id=id)

        if not self.supported_protocol_bindings:
            self.log.warning("No protocol bindings specified, "
                    "service will be invisible", service_id=self.id)

    def generate_id(self):
        return generate_message_id()

    def process(self, headers, message):

        self.log.info("Processing message", message_id=message.message_id,
                message_type=message.message_type, message_version=message.version)

        handler = self.get_message_handler(message)

        handler.validate_headers(headers, in_response_to=message.message_id)
        handler.verify_message_is_supported(message)

        try:
            response_message = handler.handle_message(self, message)
        except StatusMessageException:
            raise
        except Exception:
            raise_failure("There was a failure while executing the message handler",
                    in_response_to=message.message_id)

        if not response_message:
            raise_failure("The message handler %s did not return a TAXII Message" % handler,
                    in_response_to=message.message_id)

        return response_message

    def get_message_handler(self, message):
        try:
            return self.handlers[message.message_type]
        except KeyError:
            self.log.warning("Message not supported", message_id=message.message_id,
                    message_type=message.message_type, message_version=message.version)
            raise_failure("Message not supported by this service",
                    in_response_to=message.message_id)

    def to_service_instances(self, version):
        return service_to_service_instances(self, version)

    def get_absolute_address(self, binding):
        address = self.address

        if binding in PROTOCOL_TO_SCHEME:
            scheme = PROTOCOL_TO_SCHEME[binding]
            if scheme and not address.startswith(scheme):
                address = scheme + address
        else:
            self.log.warning("binding.not_recognized",
                             binding=binding, address=address)

        return address

    def __repr__(self):
        return "%s(id=%s, address=%s)" % (self.__class__.__name__, self.id, self.address)






import libtaxii.messages_11 as tm11
from libtaxii.constants import (
    ST_NOT_FOUND, SD_ITEM
)

from ...exceptions import StatusMessageException, raise_failure

from .base_handlers import BaseMessageHandler
from .poll_request_handlers import PollRequest11Handler, retrieve_collection


class PollFulfilmentRequest11Handler(BaseMessageHandler):

    supported_request_messages = [tm11.PollFulfillmentRequest]

    @classmethod
    def handle_message(cls, service, request):

        result_id = request.result_id
        # TODO: Forcing int pending: https://github.com/TAXIIProject/libtaxii/issues/191
        part_number = int(request.result_part_number)
        collection_name = request.collection_name

        result_set = service.get_result_set(result_id)

        collection = retrieve_collection(service, collection_name,
                in_response_to=request.message_id)

        if not result_set or result_set.collection_id != collection.id:
            raise StatusMessageException(ST_NOT_FOUND, in_response_to=request.message_id,
                status_details={SD_ITEM: result_id})
            
        response = PollRequest11Handler.prepare_poll_response(
            service = service,
            collection = collection,
            in_response_to = request.message_id, 
            timeframe = result_set.timeframe,
            content_bindings = result_set.content_bindings,
            result_part = part_number,
            allow_async = True,
            return_content = True,
            result_id = result_id
        )
        return response

# TAXII 1.0 does not support PollFulfillment


class PollFulfilmentRequestHandler(BaseMessageHandler):

    supported_request_messages = [tm11.PollFulfillmentRequest]

    @classmethod
    def handle_message(cls, service, request):
        if isinstance(request, tm11.PollFulfillmentRequest):
            return PollFulfilmentRequest11Handler.handle_message(service, request)
        else:
            raise_failure("TAXII Message not supported by message handler", request.message_id)





'''
    OpenTAXII, Python TAXII server implementation from Intelworks.
'''
# flake8: noqa

from ._version import __version__
from .server import TAXIIServer
from .config import ServerConfig
from .entities import Account

from .local import context



import json
import pytz

from opentaxii.taxii import entities

def to_collection_entity(model):
    if not model:
        return
    return entities.CollectionEntity(
        id = model.id,
        name = model.name,
        available = model.available,
        type = model.type,
        description = model.description,
        accept_all_content = model.accept_all_content,
        supported_content = deserialize_content_bindings(model.bindings)
    )


def to_block_entity(model):
    if not model:
        return

    subtypes = [model.binding_subtype] if model.binding_subtype else None

    return entities.ContentBlockEntity(
        id = model.id,
        content = model.content,

        timestamp_label = enforce_timezone(model.timestamp_label),
        content_binding = entities.ContentBindingEntity(
            model.binding_id, subtypes=subtypes),
        message = model.message,
        inbox_message_id = model.inbox_message_id,
    )


def to_inbox_message_entity(model):
    if not model:
        return

    if model.destination_collections:
        names = json.loads(model.destination_collections) 
    else:
        names = []

    return entities.InboxMessageEntity(
        id = model.id,
        message_id = model.message_id,
        original_message = model.original_message,
        content_block_count = model.content_block_count,
        destination_collections = names,

        service_id = model.service_id,

        result_id = model.result_id,
        record_count = model.record_count,
        partial_count = model.partial_count,

        subscription_collection_name = model.subscription_collection_name,
        subscription_id = model.subscription_id,

        exclusive_begin_timestamp_label = enforce_timezone(model.exclusive_begin_timestamp_label),
        inclusive_end_timestamp_label = enforce_timezone(model.inclusive_end_timestamp_label),
    )


def to_result_set_entity(model):
    if not model:
        return
    return entities.ResultSetEntity(
        result_id = model.id,
        collection_id = model.collection_id,
        content_bindings = deserialize_content_bindings(model.bindings),
        timeframe = map(enforce_timezone, (model.begin_time, model.end_time))
    )


def to_subscription_entity(model):
    if not model:
        return

    if model.params:
        parsed = dict(json.loads(model.params))
        if parsed['content_bindings']:
            parsed['content_bindings'] = deserialize_content_bindings(parsed['content_bindings'])
        params = entities.PollRequestParametersEntity(**parsed)
    else:
        params = None

    return entities.SubscriptionEntity(
        service_id = model.service_id,
        subscription_id = model.id,
        collection_id = model.collection_id,
        poll_request_params = params,
        status = model.status
    )


def to_service_entity(model):
    if not model:
        return
    return entities.ServiceEntity(id=model.id, type=model.type,
            properties=model.properties)


def serialize_content_bindings(content_bindings):
    return json.dumps([(c.binding, c.subtypes) for c in content_bindings])


def deserialize_content_bindings(content_bindings):
    raw_bindings = json.loads(content_bindings)

    bindings = []
    for (binding, subtypes) in raw_bindings:
        entity = entities.ContentBindingEntity(binding, subtypes=subtypes)
        bindings.append(entity)

    return bindings


# SQLite does not preserve TZ information
def enforce_timezone(date):

    if not date:
        return

    if date.tzinfo:
        return date

    return date.replace(tzinfo=pytz.UTC)




import json
from datetime import datetime

from sqlalchemy.orm import relationship
from sqlalchemy.schema import Table, Column, ForeignKey
from sqlalchemy.types import Integer, String, DateTime, Boolean, Text
from sqlalchemy.ext.declarative import declarative_base

__all__ = ['Base', 'ContentBlock', 'DataCollection', 'Service', 'InboxMessage', 'ResultSet', 'Subscription']

Base = declarative_base()

MAX_STR_LEN = 256


class Timestamped(Base):
    __abstract__ = True

    date_created = Column(DateTime(timezone=True), default=datetime.utcnow)

    date_updated = Column(DateTime(timezone=True), default=datetime.utcnow,
            onupdate=datetime.utcnow)


collection_to_content_block = Table('collection_to_content_block', Base.metadata,
    Column('collection_id', Integer, ForeignKey('data_collections.id')),
    Column('content_block_id', Integer, ForeignKey('content_blocks.id'))
)


class ContentBlock(Timestamped):

    __tablename__ = 'content_blocks'

    id = Column(Integer, primary_key=True)
    message = Column(Text, nullable=True)

    timestamp_label = Column(DateTime(timezone=True), default=datetime.utcnow)

    inbox_message_id = Column(Integer, ForeignKey('inbox_messages.id',
        onupdate="CASCADE", ondelete="CASCADE"), nullable=True)
    inbox_message = relationship('InboxMessage', backref='content_blocks')

    content = Column(Text)

    binding_id = Column(String(MAX_STR_LEN))
    binding_subtype = Column(String(MAX_STR_LEN))

    collections = relationship('DataCollection',
        secondary=collection_to_content_block, backref='content_blocks',
        lazy='dynamic')

    def __repr__(self):
        return 'ContentBlock(id=%s, inbox_message=%s, binding=[%s, %s])' % (
                self.id, self.inbox_message_id, self.binding_id, self.binding_subtype)


service_to_collection = Table('service_to_collection', Base.metadata,
    Column('service_id', String(MAX_STR_LEN), ForeignKey('services.id')),
    Column('collection_id', Integer, ForeignKey('data_collections.id'))
)


class Service(Timestamped):

    __tablename__ = 'services'

    id = Column(String(MAX_STR_LEN), primary_key=True)
    type = Column(String(MAX_STR_LEN))

    _properties = Column(Text, nullable=False)

    collections = relationship('DataCollection',
        secondary=service_to_collection, backref="services")

    @property
    def properties(self):
        return json.loads(self._properties)

    @properties.setter
    def properties(self, properties):
        self._properties = json.dumps(properties)


class DataCollection(Timestamped):

    __tablename__ = 'data_collections'

    id = Column(Integer, primary_key=True)
    name = Column(String(MAX_STR_LEN))

    type = Column(String(MAX_STR_LEN))
    description = Column(Text, nullable=True)

    available = Column(Boolean, default=True)
    accept_all_content = Column(Boolean, default=False)

    bindings = Column(String(MAX_STR_LEN))

    def __repr__(self):
        return u'DataCollection(%s, %s)' % (self.name, self.type)


class InboxMessage(Timestamped):

    __tablename__ = 'inbox_messages'

    id = Column(Integer, primary_key=True)

    message_id = Column(String(MAX_STR_LEN))
    result_id = Column(String(MAX_STR_LEN), nullable=True)

    record_count = Column(Integer, nullable=True)
    partial_count = Column(Boolean, default=False)

    subscription_collection_name = Column(String(MAX_STR_LEN), nullable=True)
    subscription_id = Column(String(MAX_STR_LEN), nullable=True)

    exclusive_begin_timestamp_label = Column(DateTime(timezone=True), nullable=True)
    inclusive_end_timestamp_label = Column(DateTime(timezone=True), nullable=True)

    original_message = Column(Text, nullable=False)
    content_block_count = Column(Integer)

    # FIXME: should be a proper reference ID
    destination_collections = Column(Text, nullable=True)

    service_id = Column(String(MAX_STR_LEN), ForeignKey('services.id', onupdate="CASCADE", ondelete="CASCADE"))
    service = relationship('Service', backref='inbox_messages')

    def __repr__(self):
        return 'InboxMessage(%s, %s)' % (self.message_id, self.date_created)


class ResultSet(Timestamped):

    __tablename__ = 'result_sets'

    id = Column(String(MAX_STR_LEN), primary_key=True)

    collection_id = Column(Integer, ForeignKey('data_collections.id', onupdate="CASCADE", ondelete="CASCADE"))
    collection = relationship('DataCollection', backref='result_sets')

    bindings = Column(String(MAX_STR_LEN))

    begin_time = Column(DateTime(timezone=True), nullable=True)
    end_time = Column(DateTime(timezone=True), nullable=True)


class Subscription(Timestamped):

    __tablename__ = 'subscriptions'

    id = Column(String(MAX_STR_LEN), primary_key=True)

    collection_id = Column(Integer, ForeignKey('data_collections.id', onupdate="CASCADE", ondelete="CASCADE"))
    collection = relationship('DataCollection', backref='subscriptions')

    params = Column(Text, nullable=True)

    # FIXME: proper enum type
    status = Column(String(MAX_STR_LEN))

    service_id = Column(String(MAX_STR_LEN), ForeignKey('services.id', onupdate="CASCADE", ondelete="CASCADE"))
    service = relationship('Service', backref='subscriptions')




import structlog

import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10
from libtaxii.constants import (
    SD_ITEM, ST_NOT_FOUND, ST_DENIED,
    SD_SUPPORTED_CONTENT, ST_UNSUPPORTED_CONTENT_BINDING,
    RT_FULL,
    ST_PENDING, SD_ESTIMATED_WAIT, SD_RESULT_ID, SD_WILL_PUSH
)
from libtaxii.common import generate_message_id

from .base_handlers import BaseMessageHandler
from ...exceptions import StatusMessageException, raise_failure, FailureStatus
from ....persistence.exceptions import ResultsNotReady
from ...converters import (
    content_block_entity_to_content_block, parse_content_bindings,
    content_binding_entities_to_content_bindings
)
from ...utils import get_utc_now

log = structlog.getLogger(__name__)


def retrieve_subscription(service, subscription_id, in_response_to):

    subscription = service.get_subscription(subscription_id)

    if not subscription:
        message = "The subscription requested was not found"
        details = {SD_ITEM: subscription_id}
        raise StatusMessageException(ST_NOT_FOUND, message=message,
                in_response_to=in_response_to, status_details=details)

    return subscription


def retrieve_collection(service, collection_name, in_response_to):

    collection = service.get_collection(collection_name)

    if not collection:
        message = "The collection requested was not found"
        details = {SD_ITEM: collection_name}
        raise StatusMessageException(ST_NOT_FOUND, message=message,
                in_response_to=in_response_to, status_details=details)

    if not collection.available:
        message = "The collection is not available"
        details = {SD_ITEM: collection_name}
        raise FailureStatus(message=message, in_response_to=in_response_to,
                status_details=details)

    return collection


class PollRequest11Handler(BaseMessageHandler):

    supported_request_messages = [tm11.PollRequest]

    @classmethod
    def handle_message(cls, service, request):

        if service.subscription_required and not request.subscription_id:
            message = "Subscription required"
            raise StatusMessageException(ST_DENIED, message=message,
                    in_response_to=request.message_id)

        if request.subscription_id and request.poll_parameters:
            message = "Both subscription ID and Poll Parameters present"
            log.warning(message, service_id=service.id, subscription_id=request.subscription_id)

        collection = retrieve_collection(service, request.collection_name, request.message_id)

        if request.subscription_id:
            subscription = retrieve_subscription(service, request.subscription_id,
                    request.message_id)

            if collection.id != subscription.collection_id:
                details = {SD_ITEM: request.collection_name}
                raise StatusMessageException(ST_NOT_FOUND, status_details=details,
                        in_response_to=request.message_id)

            content_bindings = subscription.params.content_bindings
            response_type = subscription.params.response_type
            allow_async = False

        else:
            params = request.poll_parameters
            raw_bindings = params.content_bindings

            requested_bindings = parse_content_bindings(raw_bindings, version=11)
            content_bindings = collection.get_matching_bindings(requested_bindings)

            if requested_bindings and not content_bindings:
                supported_bindings = content_binding_entities_to_content_bindings(
                        collection.supported_content, version=11)
                details = {SD_SUPPORTED_CONTENT: supported_bindings}
                raise StatusMessageException(ST_UNSUPPORTED_CONTENT_BINDING,
                        in_response_to=request.message_id, status_details=details)

            response_type = params.response_type
            allow_async = params.allow_asynch

        start, end = request.exclusive_begin_timestamp_label, request.inclusive_end_timestamp_label

        if (start and end) and (start > end):
            message = "Exclusive begin timestamp label is later than inclusive end timestamp label"
            raise_failure(message, request.message_id)

        return cls.prepare_poll_response(
            service = service,
            collection = collection,
            timeframe = (start, end),
            content_bindings = content_bindings,
            in_response_to = request.message_id,
            allow_async = allow_async,
            return_content = (response_type == RT_FULL),
            subscription_id = request.subscription_id
        )


    @classmethod
    def prepare_poll_response(cls, service, collection, in_response_to, timeframe=(None, None),
            content_bindings=None, result_part=1, allow_async=False, return_content=True,
            result_id=None, subscription_id=None):

        try:
            total_count = service.get_content_blocks_count(collection,
                    timeframe=timeframe, content_bindings=content_bindings)
        except ResultsNotReady:
            if not allow_async:
                message = "The content is not available now and "\
                        "the request has allow_asynch set to false"

                raise_failure(message=message, in_response_to=in_response_to)

            result_set = service.create_result_set(collection, timeframe=timeframe,
                    content_bindings=content_bindings)

            return tm11.StatusMessage(
                message_id = generate_message_id(),
                in_response_to = in_response_to,
                status_type = ST_PENDING,
                status_details = {
                    SD_ESTIMATED_WAIT: service.wait_time,
                    SD_RESULT_ID: result_set.result_id,
                    SD_WILL_PUSH: service.can_push
                }
            )

        # TODO: temporary fix, pending: https://github.com/TAXIIProject/libtaxii/issues/191
        result_part = int(result_part)

        has_more = total_count > (result_part * service.max_result_size)
        capped_count = min(service.max_result_count, total_count)
        is_partial = (capped_count < total_count)

        if has_more and not result_id:
            result_set = service.create_result_set(collection, timeframe=timeframe,
                    content_bindings=content_bindings)
            result_id = result_set.result_id

        response = tm11.PollResponse(
            message_id = generate_message_id(),
            in_response_to = in_response_to,
            collection_name = collection.name,

            more = has_more,
            result_id = result_id,
            result_part_number = result_part,

            exclusive_begin_timestamp_label = timeframe[0] if timeframe else None,
            inclusive_end_timestamp_label = timeframe[1] if timeframe else None,
            # TODO: Temporararily make capped_count an int, pending:
            #       https://github.com/TAXIIProject/libtaxii/issues/191
            record_count = tm11.RecordCount(int(capped_count), is_partial),
            subscription_id = subscription_id
        )

        if return_content:

            content_blocks = service.get_content_blocks(collection, timeframe=timeframe,
                    content_bindings=content_bindings, part_number=result_part)

            for block in content_blocks:
                response.content_blocks.append(content_block_entity_to_content_block(block, version=11))

        return response



class PollRequest10Handler(BaseMessageHandler):

    supported_request_messages = [tm10.PollRequest]

    @classmethod
    def handle_message(cls, service, request):

        collection = retrieve_collection(service, request.feed_name, request.message_id)

        if request.subscription_id:
            subscription = retrieve_subscription(service, request.subscription_id,
                    request.message_id)

            if collection.id != subscription.collection_id:
                details = {SD_ITEM: request.collection_name}
                raise StatusMessageException(ST_NOT_FOUND, status_details=details,
                        in_response_to=request.message_id)

            content_bindings = subscription.params.content_bindings
        else:
            requested_bindings = parse_content_bindings(request.content_bindings, version=10)
            content_bindings = collection.get_matching_bindings(requested_bindings)

            if requested_bindings and not content_bindings:
                supported_bindings = content_binding_entities_to_content_bindings(
                        collection.supported_content, version=10)
                details = {SD_SUPPORTED_CONTENT: supported_bindings}
                raise StatusMessageException(ST_UNSUPPORTED_CONTENT_BINDING,
                        in_response_to=request.message_id, status_details=details)

        # Only Data Feeds existed in TAXII 1.0
        if collection.type != collection.TYPE_FEED:
            message = "The Named Data Collection is not a Data Feed, it is a Data Set. " + \
                      "Only Data Feeds can be polled in TAXII 1.0"
            details = {SD_ITEM: request.feed_name}
            raise StatusMessageException(ST_NOT_FOUND, message=message,
                    status_details=details, in_response_to=request.message_id)

        start, end = request.exclusive_begin_timestamp_label, request.inclusive_end_timestamp_label

        end_response = end or get_utc_now()

        response = tm10.PollResponse(
            message_id = generate_message_id(),
            in_response_to = request.message_id,
            feed_name = collection.name,

            inclusive_begin_timestamp_label = start, #FIXME: exclusive/inclusive clash
            inclusive_end_timestamp_label = end_response,
        )

        content_blocks = service.get_content_blocks(collection, timeframe=(start, end),
                content_bindings=content_bindings)

        for block in content_blocks:
            response.content_blocks.append(content_block_entity_to_content_block(block, version=10))

        return response


class PollRequestHandler(BaseMessageHandler):

    supported_request_messages = [tm10.PollRequest, tm11.PollRequest]

    @classmethod
    def handle_message(cls, service, request):
        if isinstance(request, tm10.PollRequest):
            return PollRequest10Handler.handle_message(service, request)
        elif isinstance(request, tm11.PollRequest):
            return PollRequest11Handler.handle_message(service, request)
        else:
            raise_failure("TAXII Message not supported by message handler", request.message_id)









from opentaxii.server import TAXIIServer
from opentaxii.config import ServerConfig
from opentaxii.middleware import create_app
from opentaxii.utils import configure_logging


def run_in_dev_mode():

    config = ServerConfig()
    configure_logging(config['logging'], plain=True)

    server = TAXIIServer(config)

    app = create_app(server)
    app.debug = True

    app.run(port=9000)



import anyconfig
import argparse
import structlog

from opentaxii.config import ServerConfig
from opentaxii.server import TAXIIServer
from opentaxii.utils import configure_logging
from opentaxii.taxii.entities import CollectionEntity

config = ServerConfig()
configure_logging(config.get('logging'), plain=True)

log = structlog.getLogger(__name__)


def get_parser():
    parser = argparse.ArgumentParser(
        description = "Create services via OpenTAXII Auth API",
        formatter_class = argparse.ArgumentDefaultsHelpFormatter
    )
    return parser


def create_services():

    parser = get_parser()

    parser.add_argument("-c", "--services-config", dest="config",
            help="YAML file with services configuration", required=True)

    args = parser.parse_args()

    server = TAXIIServer(config)

    services_config = anyconfig.load(args.config, forced_type="yaml")
    server.persistence.create_services_from_object(services_config)

    log.info("Services created", count=len(services_config))


def create_collections():

    parser = get_parser()

    parser.add_argument("-c", "--collections-config", dest="config",
            help="YAML file with collections configuration", required=True)

    args = parser.parse_args()
    server = TAXIIServer(config)
    collections_config = anyconfig.load(args.config, forced_type="yaml")

    created = 0
    for collection in collections_config:

        service_ids = collection.pop('service_ids')

        existing = None
        # To keep things simple we assume here that collection name
        # is unique globally so we first check if it already exists.
        for service_id in service_ids:
            existing = server.persistence.get_collection(
                collection['name'],
                service_id)
            if existing:
                break

        if existing:
            log.warning("collection.skipped.already_exists",
                     collection_name=collection['name'],
                     existing_id=existing.id)
            continue

        entity = CollectionEntity(**collection)

        c = server.persistence.create_collection(entity)
        server.persistence.attach_collection_to_services(c.id, service_ids=service_ids)
        created += 1

    log.info("Collections created", count=created)




class OpenTAXIIPersistenceAPI(object):
    '''Abstract class that represents OpenTAXII Persistence API.

    This class defines required methods that need to exist in
    a specific Persistence API implementation.
    '''

    def create_service(self, service_entity):
        '''Create a service.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param `opentaxii.taxii.entities.ServiceEntity` service_entity:
            service entity in question
        :return: updated service entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.ServiceEntity`
        '''
        raise NotImplementedError()

    def create_collection(self, collection_entity):
        '''Create a collection.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param `opentaxii.taxii.entities.CollectionEntity` collection_entity:
            collection entity in question
        :return: updated collection entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''
        raise NotImplementedError()

    def attach_collection_to_services(self, collection_id, service_ids):
        '''Attach collection to the services.

        NOTE: Additional data management method that is not used
        in TAXII server logic but only in helper scripts.

        :param str collection_id: collection entity in question
        :param list service_ids: collection entity in question
        :return: updated collection entity, with ID field not None
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''
        raise NotImplementedError()

    def get_services(self, collection_id=None):
        '''Get the configured services.

        :param str collection_id: get only services assigned to
                                  collection with provided ID

        :return: list of service entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.ServiceEntity`
        '''
        raise NotImplementedError()

    def get_collections(self, service_id):
        '''Get the collections attached to a service.

        :param str service_id: ID of a service in question

        :return: list of collection entities.
        :rtype: list of :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''
        raise NotImplementedError()

    def get_collection(self, collection_name, service_id):
        '''Get a collection by name and service ID.

        According to TAXII spec collection name is unique per service instance.
        Method retrieves collection entity using collection name
        ``collection_name`` and service ID ``service_id`` as a composite key.

        :param str collection_name: a collection name
        :param str service_id: ID of a service

        :return: collection entity
        :rtype: :py:class:`opentaxii.taxii.entities.CollectionEntity`
        '''
        raise NotImplementedError()

    def create_inbox_message(self, inbox_message_entity): 
        '''Create an inbox message.

        :param `opentaxii.taxii.entities.InboxMessageEntity` inbox_message_entity:
            inbox message entity in question
        :return: updated inbox message entity
        :rtype: :py:class:`opentaxii.taxii.entities.InboxMessageEntity`
        '''
        raise NotImplementedError()

    def create_content_block(self, content_block_entity, collection_ids=None,
            service_id=None): 
        '''Create a content block.

        :param `opentaxii.taxii.entities.ContentBlockEntity` content_block_entity:
                content block in question
        :param list collection_ids: a list of collection IDs as strings
        :param str service_id: ID of an inbox service via which content
                block was created

        :return: updated content block entity
        :rtype: :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        '''
        raise NotImplementedError()

    def get_content_blocks_count(self, collection_id, start_time=None,
            end_time=None, bindings=None):
        '''Get a count of the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`
            
        :return: content block count
        :rtype: int
        '''
        raise NotImplementedError()

    def get_content_blocks(self, collection_id, start_time=None, end_time=None,
            bindings=None, offset=0, limit=10):
        '''Get the content blocks associated with a collection.

        :param str collection_id: ID fo a collection in question
        :param datetime start_time: start of a time frame
        :param datetime end_time: end of a time frame
        :param list bindings: list of
            :py:class:`opentaxii.taxii.entities.ContentBindingEntity`
        :param int offset: result set offset
        :param int limit: result set max size
            
        :return: content blocks list
        :rtype: list of :py:class:`opentaxii.taxii.entities.ContentBlockEntity`
        '''
        raise NotImplementedError()

    def create_result_set(self, result_set_entity):
        '''Create a result set.

        :param `opentaxii.taxii.entities.ResultSetEntity` result_set_entity:
            result set entity in question
            
        :return: updated result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        '''
        raise NotImplementedError()

    def get_result_set(self, result_set_id):
        '''Get a result set entity by ID.

        :param str result_set_id: ID of a result set.

        :return: result set entity
        :rtype: :py:class:`opentaxii.taxii.entities.ResultSetEntity`
        '''
        raise NotImplementedError()

    def create_subscription(self, subscription_entity):
        '''Create a subscription.

        :param `opentaxii.taxii.entities.SubscriptionEntity` subscription_entity:
            subscription entity in question.

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        raise NotImplementedError()

    def get_subscription(self, subscription_id):
        '''Get a subscription entity by ID.

        :param str subscription_id: ID of a subscription

        :return: subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        raise NotImplementedError()

    def get_subscriptions(self, service_id):
        '''Get the subscriptions attached to/created via a service.

        :param str service_id: ID of a service

        :return: list of subscription entities
        :rtype: list of :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        raise NotImplementedError()

    def update_subscription(self, subscription_entity):
        '''Update a subscription status.

        :param `opentaxii.taxii.entities.SubscriptionEntity` subscription_entity:
            subscription entity in question

        :return: updated subscription entity
        :rtype: :py:class:`opentaxii.taxii.entities.SubscriptionEntity`
        '''
        raise NotImplementedError()

    def get_domain(self, service_id):
        '''Get configured domain needed to create absolute URLs.

        Returns `None` by default.

        :param str service_id: ID of a service
        '''
        return None




class Account(object):
    '''Represents Account entity.

    This class holds user-specific information and is used
    for authorization.

    :param str id: account id
    :param dict details: additional details of an account
    '''

    def __init__(self, id, **details):

        self.id = id
        self.details = details
    



import sys
import structlog

from libtaxii.constants import (
        MSG_POLL_REQUEST, MSG_POLL_FULFILLMENT_REQUEST, SVC_POLL
)

from ..entities import ResultSetEntity
from .abstract import TAXIIService
from .handlers import PollRequestHandler, PollFulfilmentRequestHandler

log = structlog.getLogger(__name__)


class PollService(TAXIIService):

    handlers = {
        MSG_POLL_REQUEST : PollRequestHandler,
        MSG_POLL_FULFILLMENT_REQUEST : PollFulfilmentRequestHandler
    }
            
    service_type = SVC_POLL

    subscription_required = False
    wait_time = 300
    can_push = False

    max_result_size = sys.maxint
    max_result_count = sys.maxint

    def __init__(self, subscription_required=False, max_result_size=-1,
            max_result_count=-1, **kwargs):
        super(PollService, self).__init__(**kwargs)

        self.subscription_required = subscription_required

        self.max_result_size = max_result_size if max_result_size >= 0 else sys.maxint
        self.max_result_count = max_result_count if max_result_count >= 0 else sys.maxint


    def get_collection(self, name):
        return self.server.persistence.get_collection(name, self.id)


    def get_offset_limit(self, part_number):

        offset = (part_number - 1) * self.max_result_size
        limit = self.max_result_size

        return offset, limit


    def get_content_blocks_count(self, collection, timeframe=None,
            content_bindings=None):

        start_time, end_time = timeframe or (None, None)

        return self.server.persistence.get_content_blocks_count(
            collection_id = collection.id,
            start_time = start_time,
            end_time = end_time,
            bindings = content_bindings
        )


    def get_content_blocks(self, collection, timeframe=None, content_bindings=None,
            part_number=1):

        start_time, end_time = timeframe or (None, None)

        offset, limit = self.get_offset_limit(part_number)

        return self.server.persistence.get_content_blocks(
            collection_id = collection.id,
            start_time = start_time,
            end_time = end_time,
            bindings = content_bindings,
            offset = offset,
            limit = limit,
        )


    def create_result_set(self, collection, content_bindings=None, timeframe=(None, None)):

        result_id = self.generate_id()

        entity = ResultSetEntity(
            result_id = result_id,
            collection_id = collection.id,
            content_bindings = content_bindings,
            timeframe = timeframe
        )

        return self.server.persistence.create_result_set(entity)


    def get_result_set(self, result_set_id):
        return self.server.persistence.get_result_set(result_set_id)

    def get_subscription(self, subscription_id):
        return self.server.persistence.get_subscription(subscription_id)




from libtaxii.constants import (
    CT_DATA_FEED, CT_DATA_SET, 
    SS_ACTIVE, SS_PAUSED, SS_UNSUBSCRIBED,
    RT_FULL, RT_COUNT_ONLY
)

from .utils import is_content_supported

class Entity(object):
    '''Abstract TAXII entity class.
    '''

    def __repr__(self):
        pairs = ["%s=%s" % (k, v) for k, v in sorted(self.__dict__.items())]
        return "%s(%s)" % (self.__class__.__name__, ", ".join(pairs))


class ServiceEntity(Entity):
    '''TAXII Service entity.

    :param str type: service type, supported values are
            listed as keys in :py:attr:`opentaxii.server.TAXIIServer.TYPE_TO_SERVICE`
    :param dict properties: a dictionary with service-specific properties
    :param str id: service ID
    '''

    def __init__(self, type, properties, id=None):

        self.id = id
        self.type = type
        self.properties = properties


class ContentBindingEntity(Entity):
    '''TAXII Content Binding entity.

    :param str binding: content binding ID
    :param list subtypes: list of subtype ids
    '''

    def __init__(self, binding, subtypes=None):
        self.binding = binding
        self.subtypes = subtypes or []


class CollectionEntity(Entity):
    '''TAXII Collection entity.

    :param str id: collection id
    :param str name: collection name
    :param str description: description for the collection
    :param type: collection type
    :type type: :attr:`TYPE_FEED` or :attr:`TYPE_SET`
    :param bool accept_all_content: if collection accepts all content types
    :param list supported_content: list of the supported content bindings as
        a list of :class:`ContentBindingEntity` instances
    :param bool available: if collection is available
    '''

    TYPE_FEED = CT_DATA_FEED
    TYPE_SET = CT_DATA_SET

    def __init__(self, name, id=None, description=None, type=TYPE_FEED,
            accept_all_content=False, supported_content=None, available=True):

        self.id = id
        self.name = name
        self.available = available

        if type not in [self.TYPE_FEED, self.TYPE_SET]:
            raise ValueError('Unknown collection type "%s"' % type)

        self.type = type

        self.description = description
        self.accept_all_content = accept_all_content

        self.supported_content = []
        for content in (supported_content or []):
            if isinstance(content, basestring):
                binding = ContentBindingEntity(content)
            elif isinstance(content, tuple):
                bid, subtypes = content
                binding = ContentBindingEntity(bid, subtypes=subtypes)
            elif isinstance(content, ContentBindingEntity):
                binding = content
            else:
                raise ValueError('Unknown content binding "%s"' % content)

            self.supported_content.append(binding)


    def is_content_supported(self, content_binding):
        if self.accept_all_content:
            return True

        return is_content_supported(self.supported_content, content_binding)


    def get_matching_bindings(self, requested_bindings):
        if self.accept_all_content:
            return requested_bindings

        supported_bindings = self.supported_content

        if not supported_bindings:
            return requested_bindings

        if not requested_bindings:
            return supported_bindings

        overlap = []

        for requested in requested_bindings:
            for supported in supported_bindings:

                if requested.binding != supported.binding:
                    continue

                if not supported.subtypes:
                    overlap.append(requested)
                    continue

                if not requested.subtypes:
                    overlap.append(supported)
                    continue

                subtypes_overlap = set(supported.subtypes).intersection(requested.subtypes)

                overlap.append(ContentBindingEntity(
                    binding = requested.binding, 
                    subtypes = subtypes_overlap
                ))

        return overlap

    def __repr__(self):
        return "CollectionEntity(name=%s, type=%s, supported_content=%s)" % (
                self.name, self.type, self.supported_content)


class ContentBlockEntity(Entity):
    '''TAXII Content Block entity.

    :param str id: content block ID
    :param str content: payload
    :param datetime timestamp_label: content block timestamp
    :param `ContentBindingEntity` content_binding: content binding
    :param str message: message attached to the content block
    :param str inbox_message_id: internal ID of the inbox message entity
    '''

    def __init__(self, content, timestamp_label, content_binding=None, id=None,
            message=None, inbox_message_id=None):

        self.content = content

        self.id = id
        self.timestamp_label = timestamp_label
        self.content_binding = content_binding
        self.message = message
        self.inbox_message_id = inbox_message_id


class InboxMessageEntity(Entity):
    '''TAXII Inbox Message Entity

    :param str message_id: TAXII message ID
    :param str original_message: XML serialized original TAXII message
    :param int content_block_count: how many content blocks this message contains
    :param str service_id: ID of the Inbox Service that received the message

    :param str id: internal ID of the inbox message entity
    :param str result_id: ID of the Result Set part of which this message delivers
    :param list destination_collections: a list of destination collections, as a list of strings

    :param int record_count: how many items left in the Result Set this message is part of
    :param bool partial_count: if the record count is partial

    :param str subscription_id: ID of a subscription
    :param str subscription_collection_name: collection name of the subscription

    :param datetime exclusive_begin_timestamp_label: subscription's exclusive begin timestamp label
    :param datetime inclusive_begin_timestamp_label: subscription's inclusive begin timestamp label
    '''

    def __init__(self, message_id, original_message, content_block_count,
            service_id, id=None, result_id=None, destination_collections=None,
            record_count=None, partial_count=False, subscription_collection_name=None,
            subscription_id=None, exclusive_begin_timestamp_label=None,
            inclusive_end_timestamp_label=None):

        self.id = id

        self.message_id = message_id
        self.original_message = original_message
        self.content_block_count = content_block_count

        self.service_id = service_id

        self.destination_collections = destination_collections or []

        self.result_id = result_id
        self.record_count = record_count
        self.partial_count = partial_count

        self.subscription_collection_name = subscription_collection_name
        self.subscription_id = subscription_id

        self.exclusive_begin_timestamp_label = exclusive_begin_timestamp_label
        self.inclusive_end_timestamp_label = inclusive_end_timestamp_label



class ResultSetEntity(Entity):
    '''TAXII Result Set entity.

    :param str result_id: ID of a Result Set
    :param str collection_id: ID of a collection
    :param list content_bindings: list of :class:`ContentBindingEntity` instances
    :param tuple timeframe: a timeframe of the Result Set in a form of ``(begin, end)``
    '''

    def __init__(self, result_id, collection_id, content_bindings=None, timeframe=None):

        self.result_id = result_id

        self.collection_id = collection_id
        self.content_bindings = content_bindings or []
        self.timeframe = timeframe


class SubscriptionParameters(Entity):
    '''TAXII Subscription Parameters entity.

    Note: query formats specification is not supported

    :param str response_type: response type, supported values
        are :attr:`FULL` and :attr:`COUNT_ONLY`
    :param list content_bindings: list of :class:`ContentBindingEntity` instances
    '''

    FULL = RT_FULL
    COUNT_ONLY = RT_COUNT_ONLY

    def __init__(self, response_type=FULL, content_bindings=None):

        self.response_type = response_type
        self.content_bindings = content_bindings or []


class PollRequestParametersEntity(SubscriptionParameters):
    '''TAXII Poll Request Parameters entity.

    Note: allow_asynch and delivery_parameters fields are not supported

    :param str response_type: response type, supported values
        are :attr:`FULL` and :attr:`COUNT_ONLY`
    :param list content_bindings: list of :class:`ContentBindingEntity` instances
    '''

    def __init__(self, response_type=SubscriptionParameters.FULL, content_bindings=None):

        super(PollRequestParametersEntity, self).__init__(
                response_type=response_type, content_bindings=content_bindings)


class SubscriptionEntity(Entity):
    '''TAXII Subscription entity.

    :param str service_id: ID of a service
    :param str collection_id: ID of a collection
    :param str subscription_id: ID of a subscription
    :param str status: subscription status, supported values are:
        :attr:`ACTIVE`, :attr:`PAUSED`, :attr:`UNSUBSCRIBED`
    :param `PollRequestParametersEntity` poll_request_params:
        Poll Request Parameters entity
    '''

    ACTIVE = SS_ACTIVE
    PAUSED = SS_PAUSED
    UNSUBSCRIBED = SS_UNSUBSCRIBED

    def __init__(self, service_id, collection_id, subscription_id=None, status=ACTIVE,
            poll_request_params=None):

        self.service_id = service_id
        self.collection_id = collection_id
        self.subscription_id = subscription_id
        self.params = poll_request_params
        self.status = status






class AuthManager(object):
    '''Manager responsible for authentication.

    Manager uses API instance ``api`` for basic auth operations and
    provides additional logic on top.

    :param `opentaxii.auth.api.OpenTAXIIAuthAPI` api:
        instance of Auth API class
    '''

    def __init__(self, api):
        self.api = api

    def authenticate(self, username, password):
        '''Authenticate a user.

        :param str username: username
        :param str password: password

        :return: auth token
        :rtype: string
        '''
        return self.api.authenticate(username, password)

    def get_account(self, token):
        '''Get account for auth token.

        :param str token: auth token

        :return: an account entity
        :rtype: `opentaxii.entities.Account`
        '''
        return self.api.get_account(token)

    def create_account(self, username, password):
        '''Create an account.

        NOTE: Additional method that is only used in the helper scripts
        shipped with OpenTAXII.
        '''
        return self.api.create_account(username, password)



import os
import anyconfig

current_dir = os.path.dirname(os.path.realpath(__file__))

CONFIG_ENV_VAR = 'OPENTAXII_CONFIG'
DEFAULT_CONFIG_NAME = 'defaults.yml'
DEFAULT_CONFIG = os.path.join(current_dir, DEFAULT_CONFIG_NAME)


class ServerConfig(dict):
    '''Class responsible for loading configuration files.

    This class will load default configuration file (shipped with OpenTAXII)
    and apply user specified configuration file on top of default one.

    Users can specify custom configuration file (YAML formatted) using
    enviromental variable. The variable should contain a full path to
    a custom configuration file.

    :param str optional_env_var: name of the enviromental variable
    '''

    def __init__(self, optional_env_var=CONFIG_ENV_VAR):

        config_paths = [DEFAULT_CONFIG]

        env_var_path = os.environ.get(optional_env_var)
        if env_var_path:
            config_paths.append(env_var_path)

        options = anyconfig.load(config_paths, forced_type='yaml',
                ignore_missing=False, merge=anyconfig.MS_REPLACE)

        self.update(options)






import bcrypt

from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String
from sqlalchemy.ext.declarative import declarative_base

from werkzeug.security import safe_str_cmp

__all__ = ['Base', 'Account']

Base = declarative_base()

MAX_STR_LEN = 256


class Account(Base):

    __tablename__ = 'accounts'

    id = Column(Integer, primary_key=True)

    username = Column(String(MAX_STR_LEN), unique=True)
    password_hash = Column(String(MAX_STR_LEN))

    def set_password(self, password):
        if isinstance(password, unicode):
            password = password.encode('utf-8')
        self.password_hash = bcrypt.hashpw(password, bcrypt.gensalt())

    def is_password_valid(self, password):
        if isinstance(password, unicode):
            password = password.encode('utf-8')
        hashed = self.password_hash.encode('utf-8')
        return safe_str_cmp(bcrypt.hashpw(password, hashed), hashed)




from libtaxii.constants import (
    SVC_INBOX, MSG_INBOX_MESSAGE, SD_ACCEPTABLE_DESTINATION,
    ST_DESTINATION_COLLECTION_ERROR, ST_NOT_FOUND, SD_ITEM
)

from ..utils import is_content_supported
from ..entities import ContentBindingEntity
from ..exceptions import StatusMessageException

from ..converters import content_binding_entities_to_content_bindings, service_to_service_instances

from .abstract import TAXIIService
from .handlers import InboxMessageHandler


class InboxService(TAXIIService):

    service_type = SVC_INBOX

    handlers = {
        MSG_INBOX_MESSAGE : InboxMessageHandler
    }

    destination_collection_required = False
    accept_all_content = False
    supported_content = []


    def __init__(self, accept_all_content=False, destination_collection_required=False,
            supported_content=None, **kwargs):

        super(InboxService, self).__init__(**kwargs)

        self.accept_all_content = accept_all_content
        self.supported_content = map(ContentBindingEntity, supported_content or [])

        self.destination_collection_required = destination_collection_required


    def is_content_supported(self, content_binding, version=None):

        if self.accept_all_content:
            return True

        return is_content_supported(self.supported_content, content_binding, version=version)


    def get_destination_collections(self):
        return self.server.persistence.get_collections(self.id)


    def validate_destination_collection_names(self, name_list, in_response_to):

        name_list = name_list or []

        if (self.destination_collection_required and not name_list) or \
                (not self.destination_collection_required and name_list):

            if not name_list:
                message = 'A Destination_Collection_Name is required and none were specified'
            else:
                message = 'Destination_Collection_Names are prohibited for this Inbox Service'

            details = {SD_ACCEPTABLE_DESTINATION: [c.name for c in
                self.get_destination_collections() if c.available]}

            raise StatusMessageException(ST_DESTINATION_COLLECTION_ERROR, message=message,
                    in_response_to=in_response_to, extended_headers=details)

        if not name_list:
            return []

        collections = []

        destinations_map = dict((c.name, c) for c in self.get_destination_collections())

        for name in name_list:
            if name in destinations_map:
                collections.append(destinations_map[name])
            else:
                raise StatusMessageException(ST_NOT_FOUND, message='The Data Collection was not found',
                        in_response_to=in_response_to, extended_headers={SD_ITEM: name})

        return collections


    def to_service_instances(self, version):

        service_instances = service_to_service_instances(self, version)

        if self.accept_all_content:
            return service_instances

        for si in service_instances:
            si.inbox_service_accepted_content = self.get_supported_content(version)

        return service_instances



    def get_supported_content(self, version):

        if self.accept_all_content:
            return []

        return content_binding_entities_to_content_bindings(self.supported_content, version)




from os.path import join, dirname
from setuptools import setup, find_packages

__version__ = None
execfile('opentaxii/_version.py')

CURRENT_DIR = dirname(__file__)

def get_file_contents(filename):
    with open(join(CURRENT_DIR, filename)) as fp:
        return fp.read()

setup(
    name = 'opentaxii',
    description = 'Intelworks TAXII server implementation',
    long_description = get_file_contents('README.rst'),
    url = 'https://github.com/Intelworks/OpenTAXII',
    author = 'Intelworks',
    author_email = 'opentaxii@intelworks.com',
    version = __version__,
    license = 'BSD License',
    packages = find_packages(exclude=['tests']),
    include_package_data = True,
    package_data = {
        'opentaxii' : ['*.yml']
    },
    entry_points = {
        'console_scripts' : [
            'opentaxii-run-dev = opentaxii.cli.run:run_in_dev_mode',
            'opentaxii-create-account = opentaxii.cli.auth:create_account',
            'opentaxii-create-services = opentaxii.cli.persistence:create_services',
            'opentaxii-create-collections = opentaxii.cli.persistence:create_collections',
        ]
    },
    install_requires = [
        'pytz==2014.10',
        'pyyaml',
        'anyconfig',
        'structlog',
        'Flask',
        'sqlalchemy',
        'blinker',
        'bcrypt',
        'pyjwt',
        'libtaxii>=1.1.106'
    ],
    classifiers = [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2 :: Only',
        'Topic :: Internet',
        'Topic :: Security',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ]
)




from .base_handlers import BaseMessageHandler

import libtaxii.messages_11 as tm11
import libtaxii.messages_10 as tm10

from opentaxii.taxii.exceptions import raise_failure

from ...converters import (
    collection_to_feedcollection_information
)


class CollectionInformationRequest11Handler(BaseMessageHandler):

    supported_request_messages = [tm11.CollectionInformationRequest]

    @classmethod
    def handle_message(cls, service, request):

        response = tm11.CollectionInformationResponse(message_id=cls.generate_id(),
                in_response_to=request.message_id)

        for collection in service.advertised_collections:
            coll = collection_to_feedcollection_information(service, collection, version=11)
            response.collection_informations.append(coll)

        return response


class FeedInformationRequest10Handler(BaseMessageHandler):

    supported_request_messages = [tm10.FeedInformationRequest]

    @classmethod
    def handle_message(cls, service, request):

        response = tm10.FeedInformationResponse(message_id=cls.generate_id(),
                in_response_to=request.message_id)

        for collection in service.advertised_collections:
            feed = collection_to_feedcollection_information(service, collection, version=10)
            response.feed_informations.append(feed)

        return response


class CollectionInformationRequestHandler(BaseMessageHandler):

    supported_request_messages = [tm10.FeedInformationRequest, tm11.CollectionInformationRequest]

    @classmethod
    def handle_message(cls, service, request):

        if isinstance(request, tm10.FeedInformationRequest):
            return FeedInformationRequest10Handler.handle_message(service, request)
        elif isinstance(request, tm11.CollectionInformationRequest):
            return CollectionInformationRequest11Handler.handle_message(service, request)
        else:
            raise_failure("TAXII Message not supported by message handler", request.message_id)





import jwt
import structlog

from datetime import datetime, timedelta

from sqlalchemy import orm, engine
from sqlalchemy.orm import exc

from opentaxii.auth import OpenTAXIIAuthAPI
from opentaxii.entities import Account

from . import models

__all__ = ['SQLDatabaseAPI']


TOKEN_TTL = 60 # min

log = structlog.getLogger(__name__)


class SQLDatabaseAPI(OpenTAXIIAuthAPI):

    def __init__(self, db_connection, create_tables=False, secret=None):

        self.engine = engine.create_engine(db_connection, convert_unicode=True)

        self.Session = orm.scoped_session(orm.sessionmaker(autocommit=False,
            autoflush=True, bind=self.engine))

        attach_all(self, models)

        self.Base.query = self.Session.query_property()

        if create_tables:
            self.create_tables()

        if not secret:
            raise ValueError('Secret is not defined for %s.%s' % (
                self.__module__, self.__class__.__name__))

        self.secret = secret


    def create_tables(self):
        self.Base.metadata.create_all(bind=self.engine)


    def authenticate(self, username, password):

        try:
            account = self.Account.query.filter_by(username=username).one()
        except exc.NoResultFound:
            return

        if not account.is_password_valid(password):
            return

        return self._generate_token(account.id, ttl=TOKEN_TTL)


    def get_account(self, token):

        account_id = self._get_account_id(token)

        if not account_id:
            return

        account = self.Account.query.get(account_id)

        if not account:
            return

        return Account(id=account.id)

    def create_account(self, username, password):

        account = self.Account(username=username)
        account.set_password(password)

        session = self.Session()
        session.add(account)
        session.commit()

        return Account(id=account.id)

    def _generate_token(self, account_id, ttl=TOKEN_TTL):

        exp = datetime.utcnow() + timedelta(minutes=ttl)

        return jwt.encode({'account_id' : account_id, 'exp' : exp}, self.secret)


    def _get_account_id(self, token):
        try:
            payload = jwt.decode(token, self.secret)
        except jwt.ExpiredSignatureError:
            log.warning('Invalid token used', token=token)
            return
        except jwt.DecodeError:
            log.warning('Can not decode a token', token=token)
            return

        return payload.get('account_id')



def attach_all(obj, module):
    for key in module.__all__:
        if not hasattr(obj, key):
            setattr(obj, key, getattr(module, key))
    return obj





from libtaxii.constants import (
    VID_TAXII_XML_10, VID_TAXII_XML_11,
    VID_TAXII_SERVICES_10, VID_TAXII_SERVICES_11
)
from libtaxii.common import generate_message_id

from ...exceptions import raise_failure
from ...http import HTTP_X_TAXII_CONTENT_TYPE, HTTP_X_TAXII_SERVICES, HTTP_X_TAXII_ACCEPT


class BaseMessageHandler(object):

    supported_request_messages = []

    @staticmethod
    def generate_id():
        return generate_message_id()

    @classmethod
    def validate_headers(cls, headers, in_response_to=None):

        taxii_content_type = headers[HTTP_X_TAXII_CONTENT_TYPE]
        taxii_services = headers[HTTP_X_TAXII_SERVICES]
        taxii_accept = headers.get(HTTP_X_TAXII_ACCEPT)

        # Identify which TAXII versions the message handler supports
        supports_taxii_11 = False
        supports_taxii_10 = False

        for message in cls.supported_request_messages:
            if message.version == VID_TAXII_XML_11:
                supports_taxii_11 = True
            elif message.version == VID_TAXII_XML_10:
                supports_taxii_10 = True
            else:
                raise ValueError('The variable "supported_request_messages" '
                        'contained a non-libtaxii message module: %s' % message.__module__)

        if (taxii_services == VID_TAXII_SERVICES_11 and not supports_taxii_11) or \
                (taxii_services == VID_TAXII_SERVICES_10 and not supports_taxii_10):

            raise_failure('The specified value of %s "%s" is not supported by this TAXII Service' % (
                HTTP_X_TAXII_SERVICES, taxii_services), in_response_to)


        if (taxii_content_type == VID_TAXII_XML_11 and not supports_taxii_11) or \
                (taxii_content_type == VID_TAXII_XML_10 and not supports_taxii_10):

            raise_failure("The specified value of X-TAXII-Content-Type is not supported", in_response_to)


        if taxii_accept and \
                ((taxii_accept == VID_TAXII_XML_11 and not supports_taxii_11) or 
                (taxii_accept == VID_TAXII_XML_10 and not supports_taxii_10)):

            raise_failure("The specified value of X-TAXII-Accept is not supported", in_response_to)

        return True


    @classmethod
    def verify_message_is_supported(cls, taxii_message):
        if taxii_message.__class__ not in cls.supported_request_messages:
            raise raise_failure("TAXII Message not supported by Message Handler.", taxii_message.message_id)


    @classmethod
    def handle_message(cls, service, request):
        raise NotImplementedError()

